# -*- coding: utf-8 -*-

from .client import Client
from .oauth_client import OAuthClient
from .query_builder import QueryBuilder
from .resource import Resource
from .params_builder import ParamsBuilder

# Set default logging handler to avoid "No handler found" warnings.
import logging

try:  # Python 2.7+
    from logging import NullHandler
except ImportError:  # pragma: no cover

    class NullHandler(logging.Handler):
        def emit(self, record):
            pass


logging.getLogger(__name__).addHandler(NullHandler())



import inspect
import six
import pytz
from datetime import datetime

from .enums import (
    Boolean,
    Equality,
    DateTimeOn,
    Order,
)
from .exceptions import QueryTypeError


class Term(object):
    @staticmethod
    def wrap_constant(value, types, list_type=False):
        if list_type:
            if isinstance(value, (list, tuple)):
                if all(type(x) in types for x in value):
                    return ListValueWrapper(value, types)
                else:
                    caller = inspect.currentframe().f_back.f_code.co_name
                    raise QueryTypeError(
                        "Expected value to be a list of type %s, not %s"
                        % (types, type(value))
                    )
            else:
                caller = inspect.currentframe().f_back.f_code.co_name
                raise QueryTypeError(
                    "Invalid type passed to %s() , expected list or tuple" % (caller)
                )
        elif isinstance(value, ValueWrapper) and (
            value.type_ in types or (value.type_ == list and list_type)
        ):
            return value
        # allow other types than datetime, as long as they have strftime
        elif hasattr(value, "strftime") and datetime in types:
            return DateTimeValueWrapper(value)
        elif not type(value) in types:
            caller = inspect.currentframe().f_back.f_code.co_name
            raise QueryTypeError(
                "Invalid type passed to %s() , expected: %s" % (caller, types)
            )
        elif isinstance(value, int):
            return IntValueWrapper(value)
        elif isinstance(value, str):
            return StringValueWrapper(value)
        else:
            return value

    def eq(self, other):
        return self == other

    def gt(self, other):
        return self > other

    def gte(self, other):
        return self >= other

    def lt(self, other):
        return self < other

    def lte(self, other):
        return self <= other

    def ne(self, other):
        return self != other

    def is_empty(self):
        return IsEmptyCriterion(self)

    def is_not_empty(self):
        return NotEmptyCriterion(self)

    def is_empty_string(self):
        return IsEmptyStringCriterion(self)

    def between(self, lower, upper):
        return BetweenCriterion(
            self,
            self.wrap_constant(lower, types=[int, datetime]),
            self.wrap_constant(upper, types=[int, datetime]),
        )

    def starts_with(self, other):
        return BasicCriterion(
            "STARTSWITH", self, self.wrap_constant(other, types=[str])
        )

    def ends_with(self, other):
        return BasicCriterion("ENDSWITH", self, self.wrap_constant(other, types=[str]))

    def contains(self, other):
        return self.like(other)

    def not_contains(self, other):
        return self.not_like(other)

    def like(self, other):
        return BasicCriterion("LIKE", self, self.wrap_constant(other, types=[str]))

    def not_like(self, other):
        return BasicCriterion("NOT LIKE", self, self.wrap_constant(other, types=[str]))

    def is_in(self, other):
        return BasicCriterion(
            "IN", self, self.wrap_constant(other, types=[int, str], list_type=True)
        )

    def not_in(self, other):
        return BasicCriterion(
            "NOT IN", self, self.wrap_constant(other, types=[int, str], list_type=True)
        )

    def is_anything(self, other):
        return IsAnythingCriterion(self)

    def is_same(self, other):
        return BasicCriterion("SAMEAS", self, self.wrap_constant(other, types=[Field]))

    def is_different(self, other):
        return BasicCriterion("NSAMEAS", self, self.wrap_constant(other, types=[Field]))

    def __eq__(self, other):
        return BasicCriterion(
            Equality.eq, self, self.wrap_constant(other, types=[int, str])
        )

    def __ne__(self, other):
        return BasicCriterion(
            Equality.ne, self, self.wrap_constant(other, types=[int, str])
        )

    def __gt__(self, other):
        return BasicCriterion(
            Equality.gt, self, self.wrap_constant(other, types=[int, datetime])
        )

    def __ge__(self, other):
        return BasicCriterion(
            Equality.gte, self, self.wrap_constant(other, types=[int, datetime])
        )

    def __lt__(self, other):
        return BasicCriterion(
            Equality.lt, self, self.wrap_constant(other, types=[int, datetime])
        )

    def __le__(self, other):
        return BasicCriterion(
            Equality.lte, self, self.wrap_constant(other, types=[int, datetime])
        )

    # DateTime only
    def on(self, other):
        return DateTimeOnCriterion(
            self, self.wrap_constant(other, types=[datetime, DateTimeOn])
        )

    def not_on(self, other):
        return DateTimeNotOnCriterion(
            self, self.wrap_constant(other, types=[datetime, DateTimeOn])
        )

    # End DateTime only

    def order(self, direction):
        return OrderCriterion(self, direction)

    def __str__(self):
        return self.get_query()

    def get_query(self, **kwargs):
        raise NotImplementedError()


class Criterion(Term):
    def AND(self, other):
        return self & other

    def OR(self, other):
        return self | other

    def NQ(self, other):
        return self ^ other

    def __and__(self, other):
        return BasicCriterion(Boolean.and_, self, other)

    def __or__(self, other):
        return BasicCriterion(Boolean.or_, self, other)

    def __xor__(self, other):
        """
            While not really an XOR operation, this allows us to use Python's bitwise operators
        """
        return BasicCriterion(Boolean.nq_, self, other)

    @staticmethod
    def any(terms=()):
        crit = EmptyCriterion()

        for term in terms:
            crit |= term

        return crit

    @staticmethod
    def all(terms=()):
        crit = EmptyCriterion()

        for term in terms:
            crit &= term

        return crit

    def get_query(self, **kwargs):
        raise NotImplementedError()


class EmptyCriterion(object):
    def __and__(self, other):
        return other

    def __or__(self, other):
        return other

    def __xor__(self, other):
        return other


class BasicCriterion(Criterion):
    def __init__(self, comparator, left, right):
        """
        A wrapper for a basic criterion such as equality or inequality.
        This wraps three parts, a left and right term and a comparator which
        defines the type of comparison.


        :param comparator:
            Type: Comparator
            This defines the type of comparison, such as {quote}={quote} or
            {quote}>{quote}.
        :param left:
            The term on the left side of the expression.
        :param right:
            The term on the right side of the expression.
        """
        super(BasicCriterion, self).__init__()
        self.comparator = comparator
        self.left = left
        self.right = right

    def get_query(self, **kwargs):
        return "{left}{comparator}{right}".format(
            comparator=getattr(self.comparator, "value", self.comparator),
            left=self.left.get_query(**kwargs),
            right=self.right.get_query(**kwargs),
        )


class IsEmptyCriterion(Criterion):
    def __init__(self, term):
        super(IsEmptyCriterion, self).__init__()
        self.term = term

    def get_query(self, **kwargs):
        term = self.term.get_query(**kwargs)
        return "{}ISEMPTY".format(term)


class NotEmptyCriterion(Criterion):
    def __init__(self, term):
        super(NotEmptyCriterion, self).__init__()
        self.term = term

    def get_query(self, **kwargs):
        term = self.term.get_query(**kwargs)
        return "{}ISNOTEMPTY".format(term)


class IsAnythingCriterion(Criterion):
    def __init__(self, term):
        super(IsAnythingCriterion, self).__init__()
        self.term = term

    def get_query(self, **kwargs):
        term = self.term.get_query(**kwargs)
        return "{}ANYTHING".format(term)


class IsEmptyStringCriterion(Criterion):
    def __init__(self, term):
        super(IsEmptyStringCriterion, self).__init__()
        self.term = term

    def get_query(self, **kwargs):
        term = self.term.get_query(**kwargs)
        return "{}EMPTYSTRING".format(term)


class BetweenCriterion(Criterion):
    def __init__(self, term, start, end):
        super(BetweenCriterion, self).__init__()
        self.term = term
        self.start = start
        self.end = end

    def get_query(self, **kwargs):
        term = self.term.get_query(**kwargs)
        start = self.start.get_query(**kwargs)
        end = self.end.get_query(**kwargs)

        if isinstance(self.start, DateTimeValueWrapper) and isinstance(
            self.end, DateTimeValueWrapper
        ):
            dt_between = "%s@%s" % (start, end)
        elif isinstance(self.start, IntValueWrapper) and isinstance(
            self.end, IntValueWrapper
        ):
            dt_between = "%d@%d" % (start, end)
        else:
            raise QueryTypeError(
                "Expected both `start` and `end` of type `int` "
                "or instance of `datetime`, not %s and %s" % (type(start), type(end))
            )

        return "{}BETWEEN{}".format(term, dt_between)


class DateTimeOnCriterion(Criterion):
    def __init__(self, term, criteria):
        super(DateTimeOnCriterion, self).__init__()
        self.term = term
        self.criteria = criteria

    def get_query(self, **kwargs):
        term = self.term.get_query(**kwargs)
        if isinstance(self.criteria, DateTimeOn):
            return "{term}ON{criteria}".format(term=term, criteria=self.criteria.value)
        else:
            return "{term}ONcustom@{start}@{end}".format(
                term=term,
                start=self.criteria.get_query(date_only=True, extra_param="start"),
                end=self.criteria.get_query(date_only=True, extra_param="end"),
            )


class DateTimeNotOnCriterion(Criterion):
    def __init__(self, term, criteria):
        super(DateTimeNotOnCriterion, self).__init__()
        self.term = term
        self.criteria = criteria

    def get_query(self, **kwargs):
        term = self.term.get_query(**kwargs)
        if isinstance(self.criteria, DateTimeOn):
            return "{term}NOTON{criteria}".format(
                term=term, criteria=self.criteria.value
            )
        else:
            return "{term}NOTONcustom@{start}@{end}".format(
                term=term,
                start=self.criteria.get_query(date_only=True, extra_param="start"),
                end=self.criteria.get_query(date_only=True, extra_param="end"),
            )


class OrderCriterion(Criterion):
    def __init__(self, term, direction):
        super(OrderCriterion, self).__init__()
        self.term = term
        self.direction = direction

    def get_query(self, **kwargs):
        term = self.term.get_query(**kwargs)
        if self.direction == Order.asc or (
            isinstance(self.direction, six.string_types)
            and self.direction.lower() == "asc"
        ):
            return "ORDERBY{term}".format(term=term)
        elif self.direction == Order.desc or (
            isinstance(self.direction, six.string_types)
            and self.direction.lower() == "desc"
        ):
            return "ORDERBYDESC{term}".format(term=term)
        else:
            raise QueryTypeError(
                "Expected 'asc', 'desc', or an instance of Order, not %s"
                % (type(self.direction))
            )


class ValueWrapper(Term):
    def __init__(self, type_):
        self.type_ = type_


class IntValueWrapper(ValueWrapper):
    def __init__(self, value):
        super(IntValueWrapper, self).__init__(int)
        self.value = value

    def get_query(self, **kwargs):
        if isinstance(self.value, int):
            return self.value
        else:
            raise QueryTypeError(
                "Expected value to be an instance of `int`, not %s" % type(self.value)
            )


class StringValueWrapper(ValueWrapper):
    def __init__(self, value):
        super(StringValueWrapper, self).__init__(str)
        self.value = value

    def get_query(self, **kwargs):
        if isinstance(self.value, six.string_types):
            return self.value
        else:
            raise QueryTypeError(
                "Expected value to be an instance of `str`, not %s" % type(self.value)
            )


class DateTimeValueWrapper(ValueWrapper):
    def __init__(self, value):
        super(DateTimeValueWrapper, self).__init__(datetime)
        self.value = value

    def get_query(self, date_only=False, extra_param=None, **kwargs):
        if hasattr(self.value, "strftime"):
            datetime_ = datetime_as_utc(self.value)
            if date_only:
                value = datetime_.strftime("%Y-%m-%d")
            else:
                value = datetime_.strftime("%Y-%m-%d %H:%M:%S")

            if extra_param:
                value += '", "{}'.format(extra_param)

            return 'javascript:gs.dateGenerate("{}")'.format(value)
        else:
            raise QueryTypeError(
                "Expected value to be an instance of `datetime`, not %s"
                % type(self.value)
            )


class ListValueWrapper(ValueWrapper):
    def __init__(self, value, types):
        super(ListValueWrapper, self).__init__(list)
        self.value = value
        self.types = types

    def get_query(self, **kwargs):
        if isinstance(self.value, (list, tuple)) and all(
            type(x) in self.types for x in self.value
        ):
            return ",".join(map(str, self.value))
        else:
            raise QueryTypeError(
                "Expected value to be a list of type %s, not %s"
                % (self.types, type(self.value))
            )


class Field(Criterion):
    def __init__(self, name):
        super(Field, self).__init__()
        self.name = name

    def get_query(self, **kwargs):
        return self.name


class Table(object):
    """
        Allows the following:

        ```
        incident = Table(incident)
        criterion = incident.company.eq('3dasd3')
        ```
    """

    # Could be used to automate resource creation?
    def __init__(self, name):
        self.table_name = name

    def field(self, name):
        return Field(name)

    def __getattr__(self, name):
        return self.field(name)

    def __getitem__(self, name):
        return self.field(name)


def datetime_as_utc(date_obj):
    if date_obj.tzinfo is not None and date_obj.tzinfo.utcoffset(date_obj) is not None:
        return date_obj.astimezone(pytz.UTC)
    return date_obj



# -*- coding: utf-8 -*-

import logging
import json
import six

from .response import Response
from .exceptions import InvalidUsage

logger = logging.getLogger("pysnow")


class SnowRequest(object):
    """Creates a new :class:`SnowRequest` object.

    :param parameters: :class:`params_builder.ParamsBuilder` object
    :param session: :class:`request.Session` object
    :param url_builder: :class:`url_builder.URLBuilder` object
    """

    def __init__(
        self,
        parameters=None,
        session=None,
        url_builder=None,
        chunk_size=None,
        resource=None,
        timeout=60,
    ):
        self._parameters = parameters
        self._url_builder = url_builder
        self._session = session
        self._chunk_size = chunk_size
        self._resource = resource
        self._timeout = timeout

        self._url = url_builder.get_url()

    def _get_response(self, method, **kwargs):
        """Response wrapper - creates a :class:`requests.Response` object and passes along to :class:`pysnow.Response`
        for validation and parsing.

        :param args: args to pass along to _send()
        :param kwargs: kwargs to pass along to _send()
        :return:
            - :class:`pysnow.Response` object
        """

        params = self._parameters.as_dict()
        use_stream = kwargs.pop("stream", False)

        logger.debug(
            "(REQUEST_SEND) Method: %s, Resource: %s" % (method, self._resource)
        )

        response = self._session.request(
            method, self._url, stream=use_stream, params=params, timeout=self._timeout, **kwargs
        )
        response.raw.decode_content = True

        logger.debug(
            "(RESPONSE_RECEIVE) Code: %d, Resource: %s"
            % (response.status_code, self._resource)
        )

        return Response(
            response=response,
            resource=self._resource,
            chunk_size=self._chunk_size,
            stream=use_stream,
        )

    def _get_custom_endpoint(self, value):
        if isinstance(value, dict) and "value" in value:
            value = value["value"]
        elif not isinstance(value, six.string_types):
            raise InvalidUsage(
                "Argument 'path_append' must be a string in the following format: "
                "/path-to-append[/.../...]"
            )

        segment = value if value.startswith("/") else "/{0}".format(value)
        return self._url_builder.get_appended_custom(segment)

    def get(self, *args, **kwargs):
        """Fetches one or more records

        :return:
            - :class:`pysnow.Response` object
        """

        query = kwargs.pop("query", {}) if len(args) == 0 else args[0]

        if isinstance(query, dict):
            for key, value in query.items():
                if isinstance(value, dict):
                    query[key] = value["value"]

        self._parameters.query = query
        self._parameters.limit = kwargs.pop("limit", 10000)
        self._parameters.offset = kwargs.pop("offset", 0)
        self._parameters.fields = kwargs.pop("fields", [])
        if "display_value" in kwargs:
            self._parameters.display_value = kwargs.pop("display_value")
        if "exclude_reference_link" in kwargs:
            self._parameters.exclude_reference_link = kwargs.pop(
                "exclude_reference_link"
            )
        if "suppress_pagination_header" in kwargs:
            self._parameters.suppress_pagination_header = kwargs.pop(
                "suppress_pagination_header"
            )

        return self._get_response("GET", stream=kwargs.pop("stream", False))

    def create(self, payload):
        """Creates a new record

        :param payload: Dictionary payload
        :return:
            - Dictionary of the inserted record
        """

        return self._get_response("POST", data=json.dumps(payload))

    def update(self, query, payload):
        """Updates a record

        :param query: Dictionary, string or :class:`QueryBuilder` object
        :param payload: Dictionary payload
        :return:
            - Dictionary of the updated record
        """

        if not isinstance(payload, dict):
            raise InvalidUsage("Update payload must be of type dict")

        record = self.get(query=query).one()

        self._url = self._get_custom_endpoint(record["sys_id"])
        return self._get_response("PUT", data=json.dumps(payload))

    def delete(self, query):
        """Deletes a record

        :param query: Dictionary, string or :class:`QueryBuilder` object
        :return:
            - Dictionary containing status of the delete operation
        """

        record = self.get(query=query).one()
        self._url = self._get_custom_endpoint(record["sys_id"])

        return self._get_response("DELETE").one()

    def custom(self, method, path_append=None, **kwargs):
        """Creates a custom request

        :param method: HTTP method
        :param path_append: (optional) append path to resource.api_path
        :param headers: (optional) Dictionary of headers to add or override
        :param kwargs: kwargs to pass along to :class:`requests.Request`
        :return:
            - :class:`pysnow.Response` object
        """
        if path_append is not None:
            self._url = self._get_custom_endpoint(path_append)

        return self._get_response(method, **kwargs)



from enum import Enum


class Order(Enum):
    asc = "ASC"
    desc = "DESC"


class Comparator(Enum):
    pass


class Equality(Comparator):
    eq = "="
    ne = "!="
    gt = ">"
    gte = ">="
    lt = "<"
    lte = "<="


class Boolean(Comparator):
    and_ = "^"
    or_ = "^OR"
    nq_ = "^NQ"


# Retrieved by inspecting the `ON` dropdown
class DateTimeOn(Enum):
    today = "Today@javascript:gs.beginningOfToday()@javascript:gs.endOfToday()"
    yesterday = (
        "Yesterday@javascript:gs.beginningOfYesterday()@javascript:gs.endOfYesterday()"
    )
    tomorrow = (
        "Tomorrow@javascript:gs.beginningOfTomorrow()@javascript:gs.endOfTomorrow()"
    )
    this_week = (
        "This week@javascript:gs.beginningOfThisWeek()@javascript:gs.endOfThisWeek()"
    )
    last_week = (
        "Last week@javascript:gs.beginningOfLastWeek()@javascript:gs.endOfLastWeek()"
    )
    next_week = (
        "Next week@javascript:gs.beginningOfNextWeek()@javascript:gs.endOfNextWeek()"
    )
    this_month = (
        "This month@javascript:gs.beginningOfThisMonth()@javascript:gs.endOfThisMonth()"
    )
    last_month = (
        "Last month@javascript:gs.beginningOfLastMonth()@javascript:gs.endOfLastMonth()"
    )
    next_month = (
        "Next month@javascript:gs.beginningOfNextMonth()@javascript:gs.endOfNextMonth()"
    )
    last_3_months = "Last 3 months@javascript:gs.beginningOfLast3Months()@javascript:gs.endOfLast3Months()"
    last_6_months = "Last 6 months@javascript:gs.beginningOfLast6Months()@javascript:gs.endOfLast6Months()"
    last_9_months = "Last 9 months@javascript:gs.beginningOfLast9Months()@javascript:gs.endOfLast9Months()"
    last_12_months = "Last 12 months@javascript:gs.beginningOfLast12Months()@javascript:gs.endOfLast12Months()"
    this_quarter = "This quarter@javascript:gs.beginningOfThisQuarter()@javascript:gs.endOfThisQuarter()"
    last_quarter = "Last quarter@javascript:gs.beginningOfLastQuarter()@javascript:gs.endOfLastQuarter()"
    last_2_quarters = "Last 2 quarters@javascript:gs.beginningOfLast2Quarters()@javascript:gs.endOfLast2Quarters()"
    next_quarter = "Next quarter@javascript:gs.beginningOfNextQuarter()@javascript:gs.endOfNextQuarter()"
    next_2_quarter = "Next 2 quarters@javascript:gs.beginningOfNext2Quarters()@javascript:gs.endOfNext2Quarters()"
    this_year = (
        "This year@javascript:gs.beginningOfThisYear()@javascript:gs.endOfThisYear()"
    )
    next_year = (
        "Next year@javascript:gs.beginningOfNextYear()@javascript:gs.endOfNextYear()"
    )
    last_yesr = (
        "Last year@javascript:gs.beginningOfLastYear()@javascript:gs.endOfLastYear()"
    )
    last_2_years = "Last 2 years@javascript:gs.beginningOfLast2Years()@javascript:gs.endOfLast2Years()"
    last_7_days = "Last 7 days@javascript:gs.beginningOfLast7Days()@javascript:gs.endOfLast7Days()"
    last_30_days = "Last 30 days@javascript:gs.beginningOfLast30Days()@javascript:gs.endOfLast30Days()"
    last_60_days = "Last 60 days@javascript:gs.beginningOfLast60Days()@javascript:gs.endOfLast60Days()"
    last_90_days = "Last 90 days@javascript:gs.beginningOfLast90Days()@javascript:gs.endOfLast90Days()"
    last_120_days = "Last 120 days@javascript:gs.beginningOfLast120Days()@javascript:gs.endOfLast120Days()"
    this_hour = "Current hour@javascript:gs.beginningOfCurrentHour()@javascript:gs.endOfCurrentHour()"
    last_hour = (
        "Last hour@javascript:gs.beginningOfLastHour()@javascript:gs.endOfLastHour()"
    )
    last_2_hours = "Last 2 hours@javascript:gs.beginningOfLast2Hours()@javascript:gs.endOfLast2Hours()"
    this_minute = "Current minute@javascript:gs.beginningOfCurrentMinute()@javascript:gs.endOfCurrentMinute()"
    last_minute = "Last minute@javascript:gs.beginningOfLastMinute()@javascript:gs.endOfLastMinute()"
    last_15_minutes = "Last 15 minutes@javascript:gs.beginningOfLast15Minutes()@javascript:gs.endOfLast15Minutes()"
    last_30_minutes = "Last 30 minutes@javascript:gs.beginningOfLast30Minutes()@javascript:gs.endOfLast30Minutes()"
    last_45_minutes = "Last 45 minutes@javascript:gs.beginningOfLast45Minutes()@javascript:gs.endOfLast45Minutes()"
    one_year_ago = "One year ago@javascript:gs.beginningOfOneYearAgo()@javascript:gs.endOfOneYearAgo()"


class RelativeEquality(Comparator):
    ee = "EE"
    eq = "EE"  # on
    gt = "GT"  # after
    gte = "GE"  # on or after
    lt = "LT"  # before
    lte = "LTE"  # on or before


class RelativeTimeWindows(Enum):
    minutes = "minute"
    hours = "hour"
    days = "dayofweek"
    months = "month"
    quarters = "quarter"
    years = "year"


class RelativeDirection(Enum):
    ago = "ago"
    ahead = "ahead"
